/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.client.dataobject;

import java.io.Serializable;
import java.util.Objects;


/**
 * Basisklasse für die Spalten der Ergebnisdatensätze
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
@SuppressWarnings("serial")
public abstract class Column implements Serializable {

	/**
	 * Mögliche Typen einer Spalte.
	 *
	 * @author beck et al. projects
	 * @author Martin Hilgers
	 */
	public enum ColumnType {

		/** Reales Attribut. */REAL_ATTRIBUTE,
		/** Temporäres Attribut */TEMP_ATTRIBUTE,
		/** Zeitdauer */DURATION_ATTRIBUTE,
		/** Attributgruppe */ATTRIBUTE_GROUP
	}

	;

	/** Typ der Spalte */
	private final ColumnType _type;

	/** Referenz auf das Attribut oder die Attributgruppe, die die Werte dieser Spalte bereitstellt. */
	private final long _reference;

	/** Name der Spalte */
	private final String _columnName;

	/** Name des Elements */
	private final String _attributeName;

	/**
	 * @param type       Typ der Spalte.
	 * @param name       Name des Elements.
	 * @param reference  Referenz auf das Attribut oder die Attributgruppe, die die Werte dieser Spalte bereitstellt.
	 * @param columnName Name der Spalte. Kann <code>null</code> sein.
	 */
	public Column(ColumnType type, String name, long reference, String columnName) {
		this._type = type;
		this._attributeName = name;
		this._reference = reference;
		this._columnName = columnName;
	}

	/** @return Gibt die Referenz auf das Attribut bzw. die Attributgruppe zurück. -1 falls nicht zutreffend (z.B. Zeitdauer-Spalte, Platzhalter-Spalte) */
	public long getReference() {
		return _reference;
	}

	/** @return Returns the type. */
	public ColumnType getType() {
		return _type;
	}

	/** @return Returns the columnName. Might be <code>null</code>. */
	public String getColumnName() {
		return _columnName;
	}

	/** @return Returns the name. */
	public String getAttributeName() {
		return _attributeName;
	}

	/** @see java.lang.Object#toString() */
	@Override
    public String toString() {
		String result = _attributeName + " (" + _type + "), " + _columnName;
		return result;
	}

	/** @see java.lang.Object#equals(java.lang.Object) */
	@Override
    public boolean equals(Object object) {
		if(object == null) {
			return false;
		}
		else if(object instanceof Column) {
			Column other = (Column)object;

			return Objects.equals(this._columnName, other._columnName) && Objects.equals(this._attributeName, other._attributeName) && this._type == other._type
			       && this._reference == other._reference;
		}
		else {
			return false;
		}
	}

	/**
	 * Liefert den zugeordneten Spaltennamen oder, falls dieser <code>null</code> ist, den Namen des Elements zurück.
	 *
	 * @return Spaltenname
	 */
	public String getCaption() {
		if(_columnName != null) {
			return _columnName;
		}
		else {
			return _attributeName;
		}
	}
}
