/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.client.dataobject;

import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.pua.prot.client.dataobject.Column.ColumnType;
import de.bsvrz.pua.prot.util.attributes.AttributeGroupDescription;
import de.bsvrz.pua.prot.util.attributes.DurationAttributeDescription;
import de.bsvrz.pua.prot.util.attributes.Elements;
import de.bsvrz.pua.prot.util.attributes.RealAttributeDescription;
import de.bsvrz.pua.prot.util.attributes.TempAttributeDescription;
import de.bsvrz.sys.funclib.dataIdentificationSettings.DataIdentification;

import java.util.*;

/**
 * Modelliert den Administrationsdatensatz, der den Aufbau der Ergebnisdatensätze beschreibt
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ProtocolAdministrationData implements ProtocolResult {

	private static final long serialVersionUID = -6713427173153117154L;

	/** Falls für eine Spalte (@link #columnToDataId) keine Datenidentifikation hat. */
	public static final int NO_DID = -1;

	/** Name des Skripts */
	public String scriptName = "";

	/** Erstellungsdatum des Skripts */
	public String scriptDate = "";

	/** Versionsnummer des Skripts */
	public long version = 0;

	/** Erstellungszeitpunkt des Skripts. ms seit Epoch. */
	public long timeStamp = 0;

	/** Name des Konfigurationsverantwortlichen über den das Archivsystem angefragt wurde. */
	public String arsName = "";

	/** Spalten des Skripts. */
	public Column columns[];

	/**
	 * Zuordnung der Spalten zu Datenidentifikation. Gibt für jede Spalte an, welcher Index bei den Listen für Datenzeitstempel, Archivzeitstempel und Datenindex
	 * verwendet werden muss. Beispiel: columnToDataId[1] == 4 bedeutet, dass Spalte 2 die Datenidentifikation dids[4] verwendet. Der Datenzeitstempel für eine
	 * Spalte kann im jeweiligen Ergebnisdatensatz durch dataTimeStamps[4] (-> {@link ProtocolResultData#dataTimeStamps}) abgerufen werden. Falls einer Spalte
	 * keine Datenidentifikation zugeordnet ist, so wird der Wert {@link #NO_DID} verwendet.
	 */
	public int columnToDataId[];

	/** Liste mit den Datenidentifikationen. [x][y]: y: Id von Objekt, Attributgruppe, Aspekt */
	public long[][] dids;

	/** Simulationsvariante */
	public short simVar;

	/** Pid des Konfigurationsverantwortlichen. */
	public String configAuthPid;

	/** Ersteller des Protokolls */
	public String creatorName;

	/** Erstelldatum des Protokolls */
	public final long protocolDate;

	/** Beschreibungstext des Protokolls */
	public String infoText;

	/**
	 * @param scriptName     Name des Skripts.
	 * @param scriptDate     Erstellungsdatum des Skripts.
	 * @param version        Versionsnummer des Skripts.
	 * @param timeStamp      Erstellungszeitpunkt des Protokolls. ms seit Epoch.
	 * @param arsName        Name des Archivsystems.
	 * @param elements       Spalten des Skripts.
	 * @param columnToDataId Zurodnung der Spalten zu den Datenidentifikationen.
	 * @param dids           Liste mit Datenidentifikationen. [x][y]: y: Id von Objekt, Attributgruppe, Aspekt
	 * @param simVar         Simulationsvariante
	 * @param configAuthPid  Pid des Konfigurationsverantwortlichen.
	 * @param creatorName
	 * @param protocolDate
	 * @param infoText       Beschreibungstext des Protokolls (kann bei der Anfrage mitgegeben werden)
	 */
	public ProtocolAdministrationData(
			String scriptName,
			String scriptDate,
			long version,
			long timeStamp,
			String arsName,
			List<Elements> elements,
			int[] columnToDataId,
			long[][] dids,
			short simVar,
			String configAuthPid, final String creatorName, final long protocolDate, String infoText) {
		this.scriptName = scriptName;
		this.scriptDate = scriptDate;
		this.version = version;
		this.timeStamp = timeStamp;
		this.arsName = arsName;
		this.configAuthPid = configAuthPid;
		this.creatorName = creatorName;
		this.protocolDate = protocolDate;
		this.columns = new Column[elements.size()];
		this.columnToDataId = columnToDataId;
		this.dids = dids;
		this.simVar = simVar;
		this.infoText = infoText;
		int i = 0;
		for(Elements element : elements) {
			if(element instanceof RealAttributeDescription) {
				columns[i] = new RealAttributeColumn((RealAttributeDescription)element);
			}
			else if(element instanceof AttributeGroupDescription) {
				columns[i] = new AttributeGroupColumn((AttributeGroupDescription)element);
			}
			else if(element instanceof TempAttributeDescription) {
				columns[i] = new TempAttributeColumn((TempAttributeDescription)element);
			}
			else if(element instanceof DurationAttributeDescription) {
				columns[i] = new DurationAttributeColumn((DurationAttributeDescription)element);
			}
			else {
				columns[i] = null;
			}
			i++;
		}
	}

	/** @see java.lang.Object#equals(java.lang.Object) */
	@Override
    public boolean equals(Object object) {
		if(object == null) {
			return false;
		}
		else if(object instanceof ProtocolAdministrationData) {
			ProtocolAdministrationData other = (ProtocolAdministrationData)object;

			return Objects.equals(this.arsName, other.arsName) && Arrays.equals(this.columns, other.columns) && Objects.equals(this.scriptDate, other.scriptDate)
			       && Objects.equals(this.scriptName, other.scriptName) && this.version == other.version;
		}
		else {
			return false;
		}
	}

	/** @see java.lang.Object#toString() */
	@Override
    public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("Skript: " + scriptName + ", " + scriptDate + ", Archivsystem: " + arsName + "\nZeitstempel\t");
		for(Column c : columns) {
			if(c.getColumnName() != null && !c.getColumnName().equals("")) sb.append(c.getColumnName() + "\t");
			sb.append(c.getAttributeName() + " (");
			if(ColumnType.ATTRIBUTE_GROUP.equals(c.getType())) {
				sb.append("Attributgruppe)");
			}
			else if(ColumnType.REAL_ATTRIBUTE.equals(c.getType())) {
				sb.append("Reales Attribut)");
			}
			else if(ColumnType.TEMP_ATTRIBUTE.equals(c.getType())) {
				sb.append("Temporäres Attribut)");
			}
			else if(ColumnType.DURATION_ATTRIBUTE.equals(c.getType())) sb.append(")");
			sb.append("\t");
		}
		return sb.toString();
	}

	@Override
	public String toString(final DataModel dataModel) {
		StringBuilder sb = new StringBuilder();
		sb.append(String.format("%-28s\t", getScriptName()));
		for(Column c : columns) {
			sb.append(String.format("%28s\t", c.getAttributeName()));
		}
		sb.append("Status");
		return sb.toString();
	}

	/** @see de.bsvrz.pua.prot.client.dataobject.ProtocolResult#getType() */
	public int getType() {
		return ProtocolResult.ADMINISTRATION_DATA;
	}

	/** 
	 * Gibt den Namen des Skripts zurück
	 * @return den Namen des Skripts
	 */
	public String getScriptName() {
		return scriptName;
	}

	/** 
	 * Gibt das Erstellungs-Datum des Skripts zurück
	 * @return das Erstellungs-Datum des Skripts
	 */
	public String getScriptDate() {
		return scriptDate;
	}

	/** 
	 * Gibt die Version des Skripts zurück
	 * @return die Version
	 */
	public long getVersion() {
		return version;
	}

	/** 
	 * Gibt den Zeitstempel der Protokollerstellung zurück
	 * @return den Zeitstempel der Protokollerstellung (analog zu System.currentTimeMillis())
	 */
	public long getTimeStamp() {
		return timeStamp;
	}

	/** 
	 * Gibt den Namen des Konfigurationsverantwortlichen, über den das Archivsystem angefragt wurde zurück
	 * @return den Namen des Konfigurationsverantwortlichen, über den das Archivsystem angefragt wurde
	 */
	public String getArsName() {
		return arsName;
	}

	/** 
	 * Gibt die Anzahl Spalten zurück
	 * @return die Anzahl Spalten
	 */
	public int getColumnCount() {
		return columns.length;
	}
	
	/** 
	 * Gibt die Spalten zurück
	 * @return Liste mit Column-Objekten die eien Spalte abbilden und Informationen dazu enthalten
	 */
	public List<Column> getColumns() {
		return Collections.unmodifiableList(Arrays.asList(columns));
	}

	/**
	 * Gibt die in einer Spalte verwendete Datenidentifikation zurück
	 * @param dataModel Datenmodell
	 * @param column
	 * @return
	 */
	public DataIdentification getDataIdentification(DataModel dataModel, int column) {
		int[] columnToDataId = this.columnToDataId;
		int i = columnToDataId[column];
		if(i == NO_DID) return null;
		long[] longs = dids[i];
		return new DataIdentification(
				dataModel.getObject(longs[0]),
				new DataDescription(
						(AttributeGroup) dataModel.getObject(longs[1]),
						(Aspect) dataModel.getObject(longs[2])
				)
		);
	}

	public short getSimVar() {
		return simVar;
	}

	public String getConfigAuthPid() {
		return configAuthPid;
	}

	public String getCreatorName() {
		return creatorName;
	}

	public long getProtocolDate() {
		return protocolDate;
	}

	public String getInfoText() {
		return infoText;
	}
}
