/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.manager;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.archive.ArchiveNumQueriesResult;
import de.bsvrz.dav.daf.main.archive.ArchiveRequestManager;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.pua.prot.client.console.Console;
import de.bsvrz.pua.prot.manager.datamanager.DataManager;
import de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface;
import de.bsvrz.pua.prot.manager.status.StatePublisher;
import de.bsvrz.pua.prot.manager.taskmanager.TaskManager;
import de.bsvrz.pua.prot.manager.taskmanager.TaskManagerInterface;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.PidScript;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.kernsoftware.ConnectionManager;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;

/**
 * Submodul Manager. Ein Objekt dieser Klasse meldet sich beim Datenverteiler auf Skriptobjekte und Protokollanfragen an. Gehen Anfragen ein werden diese sofort
 * an {@link de.bsvrz.pua.prot.manager.datamanager.DataManager} oder {@link de.bsvrz.pua.prot.manager.taskmanager.TaskManager} weitergeleitet. Diese Objekte
 * werden auch vom <code>Manager</code> erzeugt.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
@SuppressWarnings("deprecation")
public class Manager {

	/** Verbindung zum Datenverteiler */
	private final ClientDavInterface _dav;

	/** Debug - Ausgaben */
	private static final Debug _debug = Debug.getLogger();

	/** Konfigurationsverantwortlicher */
	private final ConfigurationObject _configAuth;

	/** Pfad der Skript & Protokolldateien. Endet mit einem Seperator. */
	private final String _path;

	/** Anzahl der gleichzeitig erstellbaren Protokolle */
	private final int _maxProtocols;

	/** Minimal freibleibender Speicherplatz */
	private final int _quota;

	//Die folgenden Variablen werden für {@link #shutdown()} benötigt.  

	/** Datenmanager. */
	private DataManager _dataManager = null;

	/** Task-Manager. */
	private TaskManager _taskManager = null;

	/** Objekt das die Skript-bezogenen Anfragen entgegennimmt. */
	private ScriptRequestReceiver _scriptRequestReceiver;

	/** Objekt das die Protokoll-bezogenen Anfragen entgegennimmt. */
	private ProtocolRequestReceiver _protocolRequestReceiver;

	/** Telnet Server Konsole. */
	private final Console _console;

	/** Maximale Zeitdauer in ms, für die die Protokollerstellung angehalten werden kann. */
	private final long _maxDelay;

	/** Maximale Größe der Warteschlange für Anufträge zur Protokollerstellung. */
	private final int _maxQueries;
	
	/** Anzahl Minuten ohne Antwort vom Client, nach der die Übertragung getrennt wird */
	private final int _timeout;

	/**
	 * Erzeugt das Objekt.
	 *  @param dav          Verbindung zum Datenverteiler
	 * @param auth         Konfigurationsverantwortlicher
	 * @param path         Pfad der Skript- und Protokolldateien. Muss mit einem Seperator enden.
	 * @param maxProtocols Anzahl der gleichzeitig erstellbaren Protokolle
	 * @param quota        Minimal freibleibender Speicherplatz in MB.
	 * @param maxDelay     Maximale Zeitdauer in ms, für die die Protokollerstellung angehalten werden kann.
	 * @param timeout
	 * @param maxQueries   Maximale Größe der Warteschlange für Anufträge zur Protokollerstellung.
	 * @param console      Telnet Server Konsole.
	 */
	public Manager(ClientDavInterface dav, ConfigurationObject auth, String path, int maxProtocols, int quota, long maxDelay, final int timeout, int maxQueries, Console console) {
		this._dav = dav;
		_configAuth = auth;
		this._path = path;
		this._maxProtocols = maxProtocols;
		this._quota = quota;
		_timeout = timeout;
		this._console = console;
		this._maxDelay = maxDelay;
		this._maxQueries = maxQueries;
	}

	/**
	 * Startet das Submodul Manager. Führt die Anmeldungen am Datenverteiler durch. Erzeugt die Submodule Data Manager und Task Manager
	 *
	 * @param prettyPrinter Steuert ob der Pretty Printer verwendet werden soll.
	 *
	 * @throws FailureException Fehler beim Starten
	 * @see de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface
	 * @see de.bsvrz.pua.prot.manager.taskmanager.TaskManagerInterface
	 */
	public void launch(boolean prettyPrinter) throws FailureException {

		//Data Manager starten
		_debug.info("Starte Daten Manager Submodul");
		_dataManager = new DataManager(_dav, _path, _quota, prettyPrinter);
		_dataManager.start();
		_debug.info("Starte Task Manager Submodul");

		// Maximale Anzahl Anfragen mit den Einstellungen des Archivsystems vergleichen und evtl. reduzieren
		int maxProtocols = _maxProtocols;
		if(maxProtocols > 1) {
			_debug.info("Starte Abfrage nach maximaler Anzahl gleichzeitiger Archivanfragen");
			// Wenn eh nur eine Anfrage gleichzeitig ausgeführt wird, Zeit für den Abgleich mit dem Archiv-System sparen.
			// außerdem möglicher Workaround falls es Probleme bei der Ermittlung der Anzahl Anfragen gibt.
			ArchiveRequestManager archive = _dav.getArchive(_configAuth);
			if(archive.isArchiveAvailable()) {
				try {
					ArchiveNumQueriesResult numArchiveQueries = archive.getNumArchiveQueries();
					if(numArchiveQueries.isRequestSuccessful()){
						int maximumArchiveQueriesPerApplication = numArchiveQueries.getMaximumArchiveQueriesPerApplication();
						if(maximumArchiveQueriesPerApplication < maxProtocols){
							_debug.info("Das Archivsystem unterstützt nur " + maximumArchiveQueriesPerApplication
							+ " gleichzeitige Anfragen, der Wert für -maxProtokolle wird automatisch reduziert.");
							maxProtocols = maximumArchiveQueriesPerApplication;
						}
					}
					else {
						_debug.warning("Fehler bei der Ermittlung der maximalen Anzahl Archivanfragen: " + numArchiveQueries.getErrorMessage());
					}
				}
				catch(InterruptedException e) {
					_debug.warning("Fehler beim Warten auf das Archivsystem", e);
				}
				catch(IllegalStateException e) {
					_debug.warning("Fehler bei der Ermittlung der maximalen Anzahl Archivanfragen", e);
				}
			}
			else {
				_debug.warning("Fehler bei der Ermittlung der maximalen Anzahl Archivanfragen: Archivsystem nicht verfügbar");
			}
		}

		//TaskManager starten
		_taskManager = new TaskManager(_dav, _dataManager, maxProtocols, _maxQueries, _maxDelay, _timeout, _configAuth);
		//Protokollbezogene Anfragen
		subscribeForProtocolRequest(_dataManager, _taskManager);

		//Anmelden auf alle Skripte & auf die Änderung der Menge der Skriptobjekte
		subscribeForScripts(_dataManager);

		// Anmeldung von Listenern zum Protokollstatus erlauben
		StatePublisher.getInstance(_dav);
	}


	/**
	 * @param dataManager Date Manager Submodul.
	 * @param taskManager Task Manager Submodul.
	 *
	 * @throws FailureException Mehrfach-Anmeldung des gleichen Receivers
	 */
	private void subscribeForProtocolRequest(DataManager dataManager, TaskManager taskManager) throws FailureException {
		try {
			_protocolRequestReceiver = new ProtocolRequestReceiver(_dav, dataManager, taskManager, _configAuth);
		}
		catch(ConfigurationException e) {
			throw new FailureException(ErrorMessage.COMMUNICATION, e, LoggerException.ERROR);
		}
	}

	/**
	 * Meldet einen Empfänger für Skriptobjekte an. Dieser Empfänger reagiert auf alle Skriptobjekte, die in der dynamischen Menge der Skriptobjekte enthalten
	 * sind.
	 *
	 * @param dataManager Daten Manager Submodul
	 *
	 * @throws FailureException Fehler bei der Anmeldung oder beim Zugriff auf die Objektmenge.
	 */
	private void subscribeForScripts(DataManager dataManager) throws FailureException {
		try {
			DataModel model = _dav.getDataModel();
			MutableSet mutableSet = _configAuth.getMutableSet(PidScript.objectSet);
			DataDescription dd = new DataDescription(model.getAttributeGroup(PidScript.atgRequest), model.getAspect(PidScript.aspIs));
			_scriptRequestReceiver = new ScriptRequestReceiver(_dav, dataManager, mutableSet, dd, new ReceiveOptions(false), ReceiverRole.drain());
		}
		catch(ConfigurationException e) {
			throw new FailureException(ErrorMessage.COMMUNICATION, e, LoggerException.ERROR);
		}
	}

	/**
	 * Fährt die SWE herunter. Dabei werden alle Aufträge abgebrochen. Ruft abschließend {@link System#exit(int)} auf.
	 *
	 * @throws FailureException Fehler beim Herunterfahren.
	 */
	public void shutdown() throws FailureException {
		shutdownNoExit();
		System.exit(0);
	}

	/**
	 * Fährt die SWE herunter. Bricht dabei alle Aufträge ab. Ruft im Gegensatz zu {@link #shutdown()} nicht {@link System#exit(int)} auf.
	 *
	 * @throws FailureException Fehler beim Herunterfahren.
	 */
	public void shutdownNoExit() throws FailureException {
		shutdownNoExit(true);
	}

	/**
	 * Fährt die SWE herunter. Bricht dabei alle Aufträge ab. Ruft im Gegensatz zu {@link #shutdown()} nicht {@link System#exit(int)} auf.
	 *
	 * @param unsubscribe Wenn <code>true</code>, dann werden Abmeldungen beim DAV durchgeführt.
	 * @throws FailureException Fehler beim Herunterfahren.
	 */
	public void shutdownNoExit(boolean unsubscribe) throws FailureException {
		try {
			_debug.info("Protokolle und Auswertungen wird heruntergefahren.");
			if(unsubscribe && _scriptRequestReceiver != null) {
				try {
					_scriptRequestReceiver.unsubscribe();
				}
				catch(Exception e) {
					_debug.warning("Fehler beim Abmelden des scriptRequestReceivers", e);
				}
			}
			if(unsubscribe && _protocolRequestReceiver != null) {
				try {
					_protocolRequestReceiver.unsubscribe();
				}
				catch(Exception e) {
					_debug.warning("Fehler beim Abmelden des protocolRequestReceivers", e);
				}
			}
			if(_dataManager != null) {
				try {
					_dataManager.shutdown();
				}
				catch(Exception e) {
					_debug.warning("Fehler beim Beenden des dataManagers", e);
				}
			}
			if(_taskManager != null) {
				try {
					_taskManager.shutdown();
				}
				catch(Exception e) {
					_debug.warning("Fehler beim Beenden des taskManagers", e);
				}
			}
			//an sich sollte schon jetzt alles abgemeldet sein; falls eine Abmeldung jedoch nicht geklappt hat, erneut versuchen.
			if(unsubscribe) {
				try {
					ConnectionManager.unsubscribeAll(_dav);
				}
				catch(Exception e) {
					_debug.warning("Fehler beim Abmelden aller restlichen Anmeldungen", e);
				}
			}
		}
		finally {
			_debug.info("Beende Telnet Server.");
			try {
				_console.terminateServer();
			}
			catch(Exception e) {
				_debug.warning(e.getMessage(), e);
			}
			_debug.info("Herunterfahren beendet.");

			if(_taskManager != null) {
				//warten bis alle heruntergefahren
				int i = 0;
				for(; i < 5 && (!_taskManager.isTerminated()); i++) {
					_debug.warning("Es wurde noch nicht alle Threads beendet.");
					try {
						Thread.sleep(1000);
					}
					catch(InterruptedException e) {
					}
				}
				if(i == 5) throw new FailureException("Es wurde noch nicht alle Threads beendet!", LoggerException.WARNING);
			}
		}
	}


	/**
	 * Zeigt an ob das Modul und alle Submodule vollständig hochgefahren wurde.
	 *
	 * @return	<code>true</code> wenn das Modul bereit zur Annahme von Aufträgen ist. <code>false</code> sonst.
	 */
	public boolean isReady() {
		// dataManager & taskManager brauchen nicht überwacht zu werden. -> Anfragen dürfen schon eintreffen, bevor diese Submodule initialisiert wurden.
		//return dataManager.isReady() && taskManager.isReady() && scriptRequestReceiver.isReady();
		if(_scriptRequestReceiver != null) {
			return _scriptRequestReceiver.isReady();
		}
		else {
			return false;
		}
	}

	/**
	 * Liefert den Taskmanager zurück
	 *
	 * @return TaskManager. Kann <code>null</code> sein.
	 */
	public TaskManagerInterface getTaskManager() {
		return _taskManager;
	}

	/**
	 * Liefert den DataManager zurück.
	 *
	 * @return DataManager. Kann <code>null</code> sein.
	 */
	public DataManagerInterface getDataManager() {
		return _dataManager;
	}
}
