/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.manager.taskmanager;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.archive.ArchiveRequestManager;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.manager.ManagerErrorMessage;
import de.bsvrz.pua.prot.manager.ProtocolAnswer;
import de.bsvrz.pua.prot.manager.ProtocolRelatedOperation;
import de.bsvrz.pua.prot.manager.datamanager.DataManager;
import de.bsvrz.pua.prot.manager.status.StatePublisher;
import de.bsvrz.pua.prot.processing.ProcessingErrorMessage;
import de.bsvrz.pua.prot.sender.PuaSender;
import de.bsvrz.pua.prot.util.JobInProgress;
import de.bsvrz.pua.prot.util.ProcessingParameter;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.SerializerUtil;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;

import java.security.SecureRandom;
import java.util.ArrayList;
import java.util.List;

/**
 * Hauptklasse des Task Manager Submodul. Diese Klasse nimmt Anfragen vom Manager Submodul entgegen. Ist es eine Anfrage zur Protokollerstellung, so wird diese
 * einem untätigen Thread zugeordnet. Alle anderen Anfragen werden an den jeweils zuständigen Thread geleitet.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class TaskManager implements TaskManagerInterface {

	/** Verbindung zum Datenverteiler */
	private ClientDavInterface _dav;

	/** Thread-Pool */
	private ThreadPool _pool;

	/** Debug - Ausgabe */
	private static final Debug _debug = Debug.getLogger();

	/** Verbindung zum Submodul Data Manager */
	private DataManager _dataManager;

	/** Verbindung zur Konfiguration */
	private DataModel _model;

	/**
	 * Random-Objekt zur Erzeugung von Protokoll-Ids. Die Protokoll-IDs werden persitent gespeichert und sollten daher
	 * möglichst Kollisionsfrei sein. Daher werden hier Zufallszahlen verwendet. Im Falle einer (sehr unwahrscheinlichen) Kollision
	 * wird einfach eine neue Zufallszahl gebildet, siehe {@link #generateProtocolId()}.
	 */
	private final SecureRandom random = new SecureRandom();

	/**
	 * Erzeugt den Task Manager
	 *
	 * @param dav          Verbindung zum Datenverteiler
	 * @param dataManager  Daten Manager
	 * @param maxProtocols Maximale Anzahl von Threads die gleichzeitig ausgeführt werden dürfen.
	 * @param maxQueries   Maximale Größe der Warteschlange für Anufträge zur Protokollerstellung.
	 * @param maxDelay     Maximale Zeitdauer in ms, für die die Protokollerstellung angehalten werden kann.
	 * @param timeout      Anzahl Minuten ohne Antwort vom Client, nach der die Übertragung getrennt wird
	 * @param configAuth   Konfigurationsverantwortlicher für Anfragen an das Archivsystem.
	 *  @throws FailureException Keine Verbindung zum Datenverteiler; Anlegen aller Treads scheitert
	 */
	public TaskManager(final ClientDavInterface dav, final DataManager dataManager, final int maxProtocols, final int maxQueries, final long maxDelay, final int timeout, final ConfigurationObject configAuth)
			throws FailureException {
		if(dav == null) {
			_debug.error(ManagerErrorMessage.NO_DAV);
			throw new FailureException("TaskManager: " + ManagerErrorMessage.NO_DAV, LoggerException.ERROR);
		}

		this._dav = dav;
		this._model = dav.getDataModel();
		this._dataManager = dataManager;

		_pool = new ThreadPool(dav, maxProtocols, dataManager, maxQueries, maxDelay, timeout, configAuth);

		StatePublisher.getInstance(_dav).registerThreadPool(_pool, "PuA TaskManager");
	}

	/**
	 * Übergeben einer Anfrage an das Task Manager Submodul. Nun werden folgende Schritte durchgeführt:<br> 1. Feststellen, welche Operation durchgeführt werden
	 * soll<br> 2. Entpacken der Nutzdaten (aus data)<br> 3. Ausführen des Auftrages.<br>
	 *  @param client    Empfänger an den das Ergebnis der Anfrage gesendet werden soll
	 * @param id        Id der Anfrage
	 * @param operation Operationscode (gültige Werte sind {@link de.bsvrz.pua.prot.manager.ProtocolRelatedOperation}
	 * @param data      Byte-Feld mit serialisierten Anfragedaten
	 * @param initialFlowControl
	 */
	public void process(final SystemObject client, final long id, final int operation, final byte[] data, final long initialFlowControl) {
		try {
			switch(operation) {
				case ProtocolRelatedOperation.CREATE:
					createProtocol(client, id, data, initialFlowControl);
					break;
				case ProtocolRelatedOperation.ABORT:
					abortProtocolCreation(client, id, data);
					break;	
				case ProtocolRelatedOperation.ACK_FLAG:
					stopProtocolTransmission(client, id, data);
					break;
				case ProtocolRelatedOperation.STATUS:
					getStatus(client, id, data);
					break;
				case ProtocolRelatedOperation.PAUSE_PROTOCOL:
					pauseJob(data);
					break;
				case ProtocolRelatedOperation.RESUME_PROTOCOL:
					resumeJob(data);
					break;
				case ProtocolRelatedOperation.FLOW_CONTROL:
					flowControl(id, data);
					break;
				case ProtocolRelatedOperation.SEND_SAVED_PROTOCOL:
					sendSavedProtocol(client, id, data, initialFlowControl);
					break;
				case ProtocolRelatedOperation.SEND_UNREAD_PROTOCOL:
					sendUnreadProtocol(client, id, data, initialFlowControl);
					break;
				case ProtocolRelatedOperation.PUA_ARS_STATUS:
					getPuaArsStatus(client, id);
					break;
				case ProtocolRelatedOperation.JOB_LIST:
					getJobList(client, id);
					break;
				default:
					_debug.warning(ManagerErrorMessage.INVALID_OPCODE + operation);
					ProtocolAnswer.sendError(ProtocolAnswer.sender(_dav, client), id, ManagerErrorMessage.INVALID_OPCODE + operation);
					break;
			}
		}
		catch(FailureException e) {
			try {
				ProtocolAnswer.sendError(ProtocolAnswer.sender(_dav, client), id, e.getMessage());
			}
			catch(FailureException e1) {
				e1.log();
			}
			e.log();
		}
	}

	/**
	 * Setzt die Protokollerstellung oder den Protokollabruf fort.
	 *
	 * @param data Anfrageparameter
	 */
	private void resumeJob(final byte[] data) {
		ThreadElement thread = null;
		final Long jobId = SerializerUtil.deserializeId(data);
		if(jobId != null) {
			thread = _pool.getThread(jobId);
			if(thread != null) {
				thread.resumeProcessing();
			}
			else {
				_debug.finest(ManagerErrorMessage.INVALID_JOB_ID + jobId);
			}
		}
		else {
			_debug.finest(ErrorMessage.CAN_NOT_GET_JOB_ID);
		}
	}

	/**
	 * Automatische Flusskontrolle. Sorgt dafür dass Daten nur noch gesendet werden, bis die im Data-Array spezifizierte Anzahl erreicht ist.
	 *
	 * @param data Anfrageparameter
	 */
	private void flowControl(long jobId, final byte[] data) {
		ThreadElement thread = null;
		final Long numData = SerializerUtil.deserializeId(data);
		if(numData == null) return;
		thread = _pool.getThread(jobId);
		if(thread != null) {
			thread.flowControl(numData);
		}
		else {
			_debug.finest(ManagerErrorMessage.INVALID_JOB_ID + jobId + " " + numData);
		}
	}

	/**
	 * Hält die Protokollerstellung oder den Protokollabruf an.
	 *
	 * @param data Anfrageparameter
	 */
	private void pauseJob(final byte[] data) {
		ThreadElement thread = null;
		final Long jobId = SerializerUtil.deserializeId(data);
		if(jobId != null) {
			thread = _pool.getThread(jobId);
			if(thread != null) {
				thread.pauseProcessing();
			}
			else {
				_debug.info(ManagerErrorMessage.INVALID_JOB_ID + jobId);
			}
		}
		else {
			_debug.warning(ErrorMessage.CAN_NOT_GET_JOB_ID);
		}
	}

	/**
	 * Sendet den Status der Protokollerstellung an den Auftraggeber.
	 *
	 * @param client Auftraggeber
	 * @param requestId     Anfrage-Id. Entspricht NICHT der Auftrags-Id. Diese wird im Datenteil der Nachricht mitegliefert
	 * @param data   Serialisierte Auftrags-Id.
	 *
	 * @throws FailureException Fehler beim Senden.
	 */
	private void getStatus(final SystemObject client, final long requestId, final byte[] data) throws FailureException {
		ThreadElement thread = null;
		final Long jobId = SerializerUtil.deserializeId(data);
		PuaSender sender = ProtocolAnswer.sender(_dav, client);
		if(jobId != null) {
			for(ThreadPool.QueuedThread queuedThread : _pool.getWaitingProtocols()) {
				if(queuedThread.getJobId() == jobId){
					// Status 0 für Aufträge senden, die sich noch in Warteschlange befinden.
					ProtocolAnswer.sendStatus(sender, requestId, (byte)0);
					return;
				}
			}
			thread = _pool.getThread(jobId);
			if(thread != null) {
				ProtocolAnswer.sendStatus(sender, requestId, thread.getStatus());
			}
			else {
				ProtocolAnswer.sendError(sender, requestId, ManagerErrorMessage.INVALID_JOB_ID + jobId);
			}
		}
		else {
			ProtocolAnswer.sendError(sender, requestId, ErrorMessage.CAN_NOT_GET_PROTOCOL_ID);
		}
	}

	/**
	 * Sendet die Liste der aktuell in Warteschlange und in Bearbeitung befindlichen langfristigen Aufträge an den Auftraggeber.
	 *
	 * @param client Auftraggeber
	 * @param requestId     Anfrage-Id
	 *
	 * @throws FailureException Fehler beim Senden.
	 */
	private void getJobList(final SystemObject client, final long requestId) throws FailureException {
		try {
			final ArrayList<JobInProgress> jobs = _pool.getJobList();
			ProtocolAnswer.sendJobList(ProtocolAnswer.sender(_dav, client), requestId, jobs);
		}
		catch(FailureException e){
			ProtocolAnswer.sendError(ProtocolAnswer.sender(_dav, client), requestId, ErrorMessage.CAN_NOT_GET_JOB_LIST);
			_debug.warning("Fehler beim Erstellen der Auftragsliste", e);
		}
	}

	/**
	 * Sendet den Status des Systems an den Auftraggeber.
	 *
	 * @param client Auftraggeber
	 * @param id     Anfrage-Id.
	 *
	 * @throws FailureException Fehler beim Senden.
	 */
	private void getPuaArsStatus(final SystemObject client, final long id) throws FailureException {
		final ArchiveRequestManager requestManager = _dav.getArchive(_pool.getConfigAuth());
		ProtocolAnswer.sendPuaArsStatus(ProtocolAnswer.sender(_dav, client), id, requestManager.isArchiveAvailable());
	}

	/**
	 * Bricht die Protokollerstellung ab
	 *
	 * @param id   Anfrage-Id. Entspricht NICHT der Auftrags-Id. Diese wird im Datenteil der Nachricht mitgeliefert
	 * @param data Serialisierte Auftrags-Id des zu beendenden Auftrags.
	 *
	 * @throws FailureException Fehler beim Senden.
	 */
	private void abortProtocolCreation(final SystemObject client, final long id, final byte[] data) throws FailureException {
		ThreadElement thread = null;
		final Long jobId = SerializerUtil.deserializeId(data);
		PuaSender sender = ProtocolAnswer.sender(_dav, client);
		if(jobId == null) {
			ProtocolAnswer.sendError(sender, id, ErrorMessage.CAN_NOT_GET_JOB_ID);
			return;
		}
		ThreadPool.QueuedThread task = _pool.abortQueuedTask(jobId);
		if(task != null){
			// Zuerst in der Warteschlange nachsehen ob ein entsprechender Auftrag vorhanden ist und diesen ggf. Abbrechen.
			ProtocolAnswer.sendAck(sender, id);

			// Den Client, der das Protokoll angefordert hat über das Problem informieren
			ProtocolAnswer.sendAbort(ProtocolAnswer.sender(_dav, task.getClient()), task.getJobId(), ProcessingErrorMessage.ABORT_PROTOCOL);
			return;
		}
		thread = _pool.getThread(jobId);
		if(thread != null) {
			thread.abort();
			ProtocolAnswer.sendAck(sender, id);
		}
		else {
			ProtocolAnswer.sendError(sender, id, ManagerErrorMessage.INVALID_JOB_ID + jobId);
		}
	}
	
	/**
	 * Bricht die übertragung eines Protokolls ab ohne das Speichern des Protokolls abzubrechen
	 *
	 * @param id   Anfrage-Id. Entspricht NICHT der Auftrags-Id. Diese wird im Datenteil der Nachricht mitgeliefert
	 * @param data Serialisierte Auftrags-Id des zu beendenden Auftrags.
	 *
	 * @throws FailureException Fehler beim Senden.
	 */
	private void stopProtocolTransmission(final SystemObject client, final long id, final byte[] data) throws FailureException {
		ThreadElement thread = null;
		final Long jobId = SerializerUtil.deserializeId(data);
		PuaSender sender = ProtocolAnswer.sender(_dav, client);
		if(jobId == null) {
			ProtocolAnswer.sendError(sender, id, ErrorMessage.CAN_NOT_GET_JOB_ID);
			return;
		}
		ThreadPool.QueuedThread task = _pool.abortQueuedTask(jobId);
		if(task != null){
			// Zuerst in der Warteschlange nachsehen ob ein entsprechender Auftrag vorhanden ist und diesen ggf. Abbrechen.
			ProtocolAnswer.sendAck(sender, id);

			// Den Client, der das Protokoll angefordert hat über das Problem informieren
			ProtocolAnswer.sendAbort(ProtocolAnswer.sender(_dav, task.getClient()), task.getJobId(), ProcessingErrorMessage.ABORT_PROTOCOL);
			return;
		}
		thread = _pool.getThread(jobId);
		if(thread != null) {
			thread.stopProtocolTransmission();
			ProtocolAnswer.sendAck(sender, id);
		}
		else {
			ProtocolAnswer.sendError(sender, id, ManagerErrorMessage.INVALID_JOB_ID + jobId);
		}
	}

	/**
	 * Startet die Datenaufbereitung. Methode ist synchronized da in ihr die Protokoll-Id generiert wird, und dann ein neuer Thread zur Datenaufbereitung gestartet
	 * wird. Könnte diese Methode von mehreren Threads gleichzeitig aufgerufen werden, so kann die Eindeutigkeit der Protokoll-Id nicht gewährleistet werden.
	 *
	 * @param client Auftraggeber, an den das Ergebnis gesendet wird.
	 * @param id     Anfrage-Id. Ist <b>nicht</b> die Protokoll-Id, sonder eine von der Applikation vergebene Id.
	 * @param data   Anfrageparameter
	 *
	 * @param initialFlowControl
	 * @throws FailureException Fehler beim Senden.
	 */
	private synchronized void createProtocol(final SystemObject client, final long id, final byte[] data, final long initialFlowControl) throws FailureException {
		final ProcessingParameter processingParameter = ProcessingParameter.deserialize(_model, data);
		if(processingParameter == null) {
			_debug.error(ErrorMessage.CAN_NOT_DESERIALIZE);
			ProtocolAnswer.sendError(ProtocolAnswer.sender(_dav, client), id, ErrorMessage.CAN_NOT_DESERIALIZE);
		}
		else {
			final int pid = generateProtocolId();
			_pool.storeCreateQuery(id, processingParameter, client, pid, initialFlowControl);
		}
	}

	/**
	 * Sendet ein gespeichertes Protokoll an den Auftraggeber.
	 *  @param client Auftraggeber
	 * @param id     Anfrage-Id. Ist <b>nicht</b> die Protokoll-Id, sondern eine vom Auftraggeber vergebene Id.
	 * @param initialFlowControl
	 */
	private void sendSavedProtocol(final SystemObject client, final long id, final byte[] data, final long initialFlowControl) throws FailureException {
		final Long protocolId = SerializerUtil.deserializeId(data);
		if(protocolId != null) {
			_pool.storeGetProtocolQuery(id, client, protocolId, initialFlowControl);
		}
		else {
			throw new FailureException(ErrorMessage.UNEXPECTED_DATA_TYPE + SerializerUtil.deserialize(data) + " <-> Long", LoggerException.WARNING);
		}
	}

	/**
	 * Sendet ein gespeichertes Protokoll an den Auftraggeber.
	 *  @param client Auftraggeber
	 * @param id     Anfrage-Id. Ist <b>nicht</b> die Protokoll-Id, sondern eine vom Auftraggeber vergebene Id.
	 * @param initialFlowControl
	 */
	private void sendUnreadProtocol(final SystemObject client, final long id, final byte[] data, final long initialFlowControl) throws FailureException {
		final Long protocolId = SerializerUtil.deserializeId(data);
		if(protocolId != null) {
			_pool.storeGetUnreadProtocolQuery(id, client, protocolId, initialFlowControl);
		}
		else {
			throw new FailureException(ErrorMessage.UNEXPECTED_DATA_TYPE + SerializerUtil.deserialize(data) + " <-> Long", LoggerException.WARNING);
		}
	}

	/**
	 * Generiert eine (global) eindeutige Protokoll-Id
	 *
	 * @return Protokoll-Id
	 */
	private int generateProtocolId() {
		int pid;
		while(true) {
			pid = random.nextInt() & 0x7fffffff; // Vorzeichen-Bit entfernen
			if(_pool.getThread(pid) == null && _dataManager.isUnique(pid)) {
				break;
			}
			if(pid < 0x10000000){
				// Protocol-IDs und Anfrage-IDs werden unabhängig von Client und von Server generiert.
				// Da es dabei potentiell zu problematischen Kollisionen kommen kann, werden sehr kleine Protokoll-IDs
				// hier verworfen und die Kollisionswahrscheinlichkeit zu reduzieren.
				// (Der PuA-Client zählt die IDs beginnende bei 1 hoch)
				// Es ist unwahrscheinlich, dass ein Client mehr als 200 Mio. Anfragen erzeugt und für den Fall ist die Kollisions
				// wahrscheinlichkeit immer noch geringer als 1:200 Mio., d.h. das Risiko ist vertretbar.
				// Neuere Clients beschränken die Request-ID automatisch auf einen Bereich zwischen 0 und 0x10000000.
				break;
			}
		}
		return pid;
	}

	/** @see de.bsvrz.pua.prot.manager.taskmanager.TaskManagerInterface#shutdown() */
	public void shutdown() {
		_pool.killAll();
	}

	/**
	 * Zeigt an ob der Taskmanager und alle Subthreads beendet wurden.
	 *
	 * @return <code>true</code> falls der Taskmanager beendet wurde.
	 */
	public boolean isTerminated() {
		return _pool.isTerminated();
	}

	/** @see de.bsvrz.pua.prot.manager.taskmanager.TaskManagerInterface#getActiveProtocols() */
	public Long[] getActiveProtocols() {
		return _pool.getActiveProtocols();
	}

	/** @see de.bsvrz.pua.prot.manager.taskmanager.TaskManagerInterface#getWaitingProtocols() */
	public List<ThreadPool.QueuedThread> getWaitingProtocols() {
		return _pool.getWaitingProtocols();
	}

	@Override
	public ThreadPool getPool() {
		return _pool;
	}
}
