/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.processing.util;

import de.bsvrz.sys.funclib.losb.util.Tuple;

import java.util.*;


/**
 * Balancierter binärer Suchbaums mit fester Knotenanzahl. Ermöglicht die effiziente Feststellung, ob ein Zeitstempel innerhalb der Zeitintervalle liegt, die in
 * den Blättern des Knoten gespeichert sind.<br> Das Hinzufügen oder entfernen Knoten ist nicht möglich.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class BinaryTree {

	/** Binärer Suchbaum. Ist <ocde>null</code> falls <code>periods</code> leer ist. */
	private BinaryTreeNode _binaryTree;

	/** Liste mit Zeitintervallen. */
	private List<Tuple<Long, Long>> _periods;

	/**
	 * Erzeugt einen binären Suchbaum.
	 *
	 * @param periods Zeitintervalle. Liste wird nicht verändert.
	 */
	public BinaryTree(List<Tuple<Long, Long>> periods) {
		this._periods = periods;
		_binaryTree = BinaryTreeNode.createTree(periods);
	}

	/**
	 * Liefert das Anfang des Zeitintervalls <code>index</code> zurück.
	 *
	 * @param index Index des Zeitintervalls
	 *
	 * @return Ende des Zeitintervalls oder -1, falls ein ungültiger Index verwendet wurde oder keine Zeitstempel verwaltet werden.
	 */
	public long getStartOfInterval(int index) {
		if(_periods.size() > index) {
			return _periods.get(index).first;
		}
		else {
			return -1;
		}
	}

	/**
	 * Liefert das Ende des Zeitintervalls <code>index</code> zurück.
	 *
	 * @param index Index des Zeitintervalls
	 *
	 * @return Ende des Zeitintervalls oder -1, falls ein ungültiger Index verwendet wurde oder keine Zeitstempel verwaltet werden.
	 */
	public long getEndOfInterval(int index) {
		if(_periods.size() > index) {
			return _periods.get(index).last;
		}
		else {
			return -1;
		}
	}

	/**
	 * Liefert das Ende aller Zeitintervalle zurück. Es wird der größten Zeitstempel zurückgeliefert, der vom Baum verwaltet wird.
	 *
	 * @return Ende aller Zeitintervalls oder -1, falls ein ungültiger Index verwendet wurde oder keine Zeitstempel verwaltet werden.
	 */
	public long getEndOfAllIntervalls() {
		if(_binaryTree != null) {
			return _binaryTree.getMaxTimeStamp();
		}
		else {
			return -1;
		}
	}

	/**
	 * Anzahl der verwalteten Zeitintervalle.
	 *
	 * @return Anzahl der im Baum gespeicherten Zeitintervalle.
	 */
	public int size() {
		return _periods.size();
	}

	/**
	 * Prüft, ob ein Zeitstempel in den gespeicherten Intervallen liegt.
	 *
	 * @param timeStamp Zeitstempel.
	 *
	 * @return	<code>true</code>, falls der Zeitstempel innerhalb der gespeicherten Zeitintervalle liegt. <code>false</code> sonst.
	 */
	public boolean contains(long timeStamp) {
		if(_binaryTree != null) {
			return _binaryTree.contains(timeStamp);
		}
		else {
			return false;
		}
	}


	/** @see java.lang.Object#toString() */
	@Override
    public String toString() {
		return _binaryTree.toString();
	}

	/**
	 * Zeigt ob Zeitintervalle im Baum gespeichert sind.
	 *
	 * @return	<code>true</code> falls Zeitintevalle im Baum gespeichert sind.
	 */
	public boolean isEmpty() {
		return _periods.isEmpty();
	}
}

/**
 * Knoten des balancierten binären Suchbaums mit fester Knotenanzahl. Ermöglicht die effiziente Feststellung, ob ein Zeitstempel innerhalb der Zeitintervalle
 * liegt, die in den Blättern des Knoten gespeichert sind. <br> Zum erzeugen des binären Suchbaums muss die Methode {@link #createTree(List)} verwendet werden.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
class BinaryTreeNode {

	/** Größter Zeitstempel des linken Pfades dieses Knotens. */
	private long leftMaxTimeStamp;

	/** Größter Zeitstempel, der in den Kindern dieses Knotens gespeichert wird. */
	private long maxTimeStamp;

	/** Linkes Kind */
	private BinaryTreeNode left;

	/** rechtes Kind */
	private BinaryTreeNode right;


	/**
	 * Erzeugt die Baumstruktur für die übergebenen Zeitintervalle.
	 *
	 * @param periods Zeitintervalle. Die Zeitintervalle müssen aufsteigend sortiert sein, und dürfen einander nicht überschneiden.
	 *
	 * @return Balancierter, binärer Suchbaum. Falls kein Zeitintervall übergeben wurde, wird null zurückgegeben.
	 */
	public static BinaryTreeNode createTree(List<Tuple<Long, Long>> periods) {
		if(periods.size() == 0) {
			return null;
		}
		else {
			//Nodes enthält alle erzeugten Knoten der untersten Ebene (in dieser Methode)
			ArrayList<BinaryTreeNode> nodes = new ArrayList<BinaryTreeNode>(periods.size() / 2 + 1);

			int i = 0;
			//Einem Knoten immer zwei Zeitintervalle als Kinder zuordnen
			while(i + 1 < periods.size())		//solange laufen, bis ein oder kein Zeitintervall übrig bleibt
			{
				//Einen Knoten für je zwei Zeitintervalle erzeugen
				nodes.add(new BinaryTreeNode(new BinaryTreeLeaf(periods.get(i)), new BinaryTreeLeaf(periods.get(i + 1))));
				i += 2;
			}

			//Ein einzeler Zeitbereich wird in einem BinaryTreeLeaf gespeichert
			if(i == periods.size() - 1)		//Anzahl der Zeitintervalle war ungerade -> ein einzelnes Kind
			{
				nodes.add(new BinaryTreeLeaf(periods.get(i)));
			}

			//jetzt falls nötig Baumstruktur aufbauen:
			if(nodes.size() >= 2)			//Es gibt mehr als zwei Knoten -> Vaterknoten erzeugen
			{
				return buildTree(nodes);
			}
			else {
				return nodes.get(0);
			}
		}
	}


	/**
	 * Erzeugt einen Baum aus den Knoten.
	 *
	 * @param nodes Knoten der untersten Ebene, d.h. an die Kinder dieser Knoten sind Blätter. Es muss in der Liste mindestens zwei Knoten geben.
	 *
	 * @return Balancierter Binärbaum
	 */
	private static BinaryTreeNode buildTree(ArrayList<BinaryTreeNode> nodes) {
		for(int i = 0; i + 1 < nodes.size(); i++)		//i+1<size, da es immer zwei Knoten geben muss!
		{
			// In jedem Schleifendurchlauf wird die Anzahl der Knoten um eins verringert
			BinaryTreeNode node = nodes.remove(i);	//Knoten entfernen
			nodes.set(i, new BinaryTreeNode(node, nodes.get(0)));	//zwei Knoten mergen
		}
		if(nodes.size() == 1) {
			return nodes.get(0);
		}
		else {
			return buildTree(nodes);	//weitere Ebene mit Vaterknoten hinzufügen
		}
	}

	/**
	 * Setzt das linke und rechte Kind des Knotens
	 *
	 * @param left  Linkes Kind
	 * @param right Rechtes Kind
	 */
	protected BinaryTreeNode(BinaryTreeNode left, BinaryTreeNode right) {
		this.left = left;
		this.leftMaxTimeStamp = left.getMaxTimeStamp();

		this.right = right;
		this.maxTimeStamp = right.getMaxTimeStamp();
	}

	/** Konstruktor mit minimaler Initialisierung. */
	BinaryTreeNode() {
		left = null;
		right = null;
		maxTimeStamp = -1;
		leftMaxTimeStamp = -1;
	}


	/**
	 * Prüft, ob ein Zeitstempel in den gespeicherten Intervallen liegt.
	 *
	 * @param timeStamp Zeitstempel.
	 *
	 * @return	<code>true</code>, falls der Zeitstempel innerhalb der gespeicherten Zeitintervalle liegt. <code>false</code> sonst.
	 */
	public boolean contains(long timeStamp) {
		if(timeStamp <= leftMaxTimeStamp) {
			return left.contains(timeStamp);
		}
		else {
			return right.contains(timeStamp);
		}
	}

	/**
	 * Liefert den größten Zeitstempel zurück, der in diesem Knoten gespeichert ist.
	 *
	 * @return Größter Zeitstempel.
	 */
	public long getMaxTimeStamp() {
		return maxTimeStamp;
	}


	/** @see java.lang.Object#toString() */
	@Override
    public String toString() {
		return "(" + ((left != null) ? left.toString() : "") + ((right != null) ? right.toString() : "") + ")";
	}
}

/**
 * Blatt des binären Suchbaums. Speichert ein Zeitintervall.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 */
class BinaryTreeLeaf extends BinaryTreeNode {

	/** Start des Intervalls */
	private long start;

	/** Ende des Intervalls */
	private long end;

	/**
	 * Erzeugt ein Blatt, das ein Zeitintervall speichert.
	 *
	 * @param period Zeitintervall
	 */
	public BinaryTreeLeaf(Tuple<Long, Long> period) {
		super();
		this.start = period.first;
		this.end = period.last;
	}

	/** @see de.bsvrz.pua.prot.processing.util.BinaryTreeNode#contains(long) */
	@Override
    public boolean contains(long timeStamp) {
		return timeStamp >= start && timeStamp <= end;
	}

	/** @see de.bsvrz.pua.prot.processing.util.BinaryTreeNode#getMaxTimeStamp() */
	@Override
    public long getMaxTimeStamp() {
		return end;
	}

	/** @see java.lang.Object#toString() */
	@Override
    public String toString() {
		return "(" + start + "," + end + ")";
	}
}
