/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.processing.util;

import de.bsvrz.pua.prot.util.ProcessingInformation;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;

/**
 * Ausgangsdatensatz. Basis des Ergebnisdatensatzes {@link de.bsvrz.pua.prot.client.dataobject.ProtocolResult}
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class IntermediateDataSet {

	/** Wird ins Statusbyte geschrieben. Zeigt an, dass es sich um einen Datensatz handelt, der dem vorherigen gleicht. */
	public static final byte NO_CHANGE = -5;

	private List<BaseDataSet> _baseDataSets;

	private byte _status;

	private ProcessingInformation.ApplyAggregation _order;

	/** Zeitstempel des Elements mit dem kleinsten nicht aufgefüllten Datenzeitstempel. */
	private long _winnerTimeStamp;

	/** Index zu welchem Zeitintervall der Datensatz gehört */
	private int _intervalIndex;


	/**
	 * @param baseDataSets    Verwendete Elemente im Ausgangsdatensatz.
	 * @param status          Status der Protokollerstellung
	 * @param order           Aggregation, die diesen Ausgangsdatensatz erzeugt hat.
	 * @param winnerTimeStamp Zeitstempel des Elements in <code>baseDataSets</code> mit dem kleinsten nicht aufgefüllten Datenzeitstempel.
	 * @param intervalIndex   Aktueller Zeitbereich
	 */
	public IntermediateDataSet(
			List<BaseDataSet> baseDataSets, byte status, ProcessingInformation.ApplyAggregation order, long winnerTimeStamp, final int intervalIndex) {
		this._baseDataSets = baseDataSets;
		this._status = status;
		this._order = order;
		this._winnerTimeStamp = winnerTimeStamp;
		_intervalIndex = intervalIndex;
	}

	/**
	 * @param baseDataSets Verwendete Elemente im Ausgangsdatensatz.
	 * @param status       Status der Protokollerstellung
	 * @param order        Aggregation, die diesen Ausgangsdatensatz erzeugt hat.
	 * @param intervalIndex Aktueller Zeitbereich
	 */
	public IntermediateDataSet(List<BaseDataSet> baseDataSets, byte status, ProcessingInformation.ApplyAggregation order, final int intervalIndex) {
		this(baseDataSets, status, order, 0, intervalIndex);
	}

	/**
	 * @param baseDataSets    Verwendete Elemente im Ausgangsdatensatz.
	 * @param status          Status der Protokollerstellung
	 * @param order           Aggregation, die diesen Ausgangsdatensatz erzeugt hat.
	 * @param winnerTimeStamp Zeitstempel des Elements in <code>baseDataSets</code> mit dem kleinsten nicht aufgefüllten Datenzeitstempel.
	 * @param intervalIndex Aktueller Zeitbereich
	 */
	public IntermediateDataSet(
			final BaseDataSet[] baseDataSets,
			final byte status,
			final ProcessingInformation.ApplyAggregation order,
			final long winnerTimeStamp,
			final int intervalIndex) {
		this(Arrays.asList(baseDataSets),status,order,winnerTimeStamp,intervalIndex);
	}

	/** @see java.lang.Object#toString() */
	@Override
    public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("Zeitstempel: ").append(Util.msToDate(_winnerTimeStamp)).append("(").append(_winnerTimeStamp).append("), erzeugt durch ").append(_order).append(
				", Status: "
		).append(_status).append("%");
		for(BaseDataSet bds : _baseDataSets) {
			sb.append(", ").append(bds);
		}
		return sb.toString();
	}

	/**
	 * Zeitstempel des Elements mit dem kleinsten nicht aufgefüllten Datenzeitstempel.
	 *
	 * @return Zeitstempel des Ausgangsdatensatzes.
	 */
	public long getWinnerTimeStamp() {
		return _winnerTimeStamp;
	}

	/** Aggregation, die diesen Ausgangsdatensatz erzeugt hat. */
	public ProcessingInformation.ApplyAggregation getOrder() {
		return _order;
	}

	/** Status der Protokollerstellung */
	public byte getStatus() {
		return _status;
	}

	/** Verwendete Elemente im Ausgangsdatensatz.
	 * @return Liste Mis BastDataSets
	 */
	public List<BaseDataSet> getBaseDataSets() {
		return Collections.unmodifiableList(_baseDataSets);
	}

	public int getIntervalIndex() {
		return _intervalIndex;
	}
}
