/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.util;

import antlr.collections.AST;
import de.bsvrz.pua.prot.exceptions.SemanticErrorException;
import de.bsvrz.pua.prot.interpreter.InterpreterErrorMessage;
import de.bsvrz.pua.prot.processing.util.BaseDataSet;
import de.bsvrz.pua.prot.processing.util.ExpressionResultAndState;
import de.bsvrz.pua.prot.processing.util.TempAttribut;
import de.bsvrz.pua.prot.util.ExpressionResult.ResultType;
import de.bsvrz.pua.prot.util.attributes.AttributeDescription;
import de.bsvrz.pua.prot.util.attributes.RealAttributeDescription;
import de.bsvrz.pua.prot.util.attributes.RealElement;
import de.bsvrz.pua.prot.util.attributes.TempAttributeDescription;

import java.util.*;

import static de.bsvrz.pua.prot.processing.util.ExpressionResultAndState.deriveResult;


/**
 * Modelliert einen binären Baum von Ausdrücken. Es ist möglich, dass ein Knoten nur ein Kind hat (für unäre Operatoren) Die Knoten sind Operatoren, die Blätter
 * Zeichenketten, Bool, Attribute oder Zahlen. <br> OPT Methode implementieren, die den Baum vereinfacht, d.h. alle nicht variablen Knoten bereits auswertet (->
 * kürzere Wege)
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ExpressionTree {

	/**
	 * Im Ausdruck mögliche Operationen.
	 *
	 * @author beck et al. projects GmbH
	 * @author Martin Hilgers
	 */
	public enum Operation {

		/** Negation */neg(false, "-"),
		/** Not */not(false, "Nicht "),
		/** Multiplikation */mult(true, " * "),
		/** Division */div(true, " / "),
		/** Integer Division */divInt(true, " div "),
		/** Modula */modulo(true, " modulo "),
		/** Plus */add(true, " + "),
		/** Minus */subtract(true, " - "),
		/** String Konkatenation */conc(true, " & "),
		/** Kleiner */less(true, " < "),
		/** Kleiner gleich */lessEqual(true, " <= "),
		/** Größer */greater(true, " > "),
		/** Größer gleich */greaterEqual(true, " >= "),
		/** Gleich */equal(true, " = "),
		/** Ungleich */notEqual(true, " <> "),
		/** Und */and(true, " und "),
		/** Oder */or(true, " oder "),
		/** Ungültige Operation */invalid(false, " <Ungültig> ");

		private final boolean _binary;
		private final String _symbol;

		Operation(boolean binary, String symbol) {
			_binary = binary;
			_symbol = symbol;
		}

		public boolean isBinary() {
			return _binary;
		}

		public String getSymbol() {
			return _symbol;
		}
	}

	;

	/** Linker Operand */
	public ExpressionTree left;

	/** Rechter Operand. */
	public ExpressionTree right;

	/** Operator. */
	public Operation operation;

	/**
	 * Erzeugt einen Ausdruck.
	 *
	 * @param l linker Operand.
	 * @param r rechter Operand.
	 * @param o Operator.
	 */
	public ExpressionTree(final ExpressionTree l, final ExpressionTree r, final Operation o) {
		left = l;
		right = r;
		operation = o;
	}

	/** Ausdruck ohne Operanden und <code>invalid</code> Operand. */
	public ExpressionTree() {
		left = null;
		right = null;
		operation = Operation.invalid;
	}

	/**
	 * Liefert alle Attribute, auf die der Baum zugreift, zurück. Liefert erst alle Attribute aus dem linken Teilbaum des Baumes. Anschließend werden die Objekte
	 * aus dem rechten Teilbaum zurückgegeben. Wird ein Attribut mehrfach verwendet, so wird nur das erste Auftreten berücksichtigt.
	 *
	 * @return Attribute, die im Ausdruck verwendet werden. Leeres Liste, falls keine Attribute verwendet werden.
	 */
	public List<AttributeDescription> getAttributes() {
		final ArrayList<AttributeDescription> attributes = new ArrayList<AttributeDescription>();
		getAttributeOrder(attributes, new HashSet<AttributeDescription>());
		return attributes;
	}

	/**
	 * Liefert alle Attribute auf die der Baum zugreift. Dublikate werden verworfen, d.h. jedes Attribut wird nur einmal zurückgeliefert.
	 *
	 * @param list    liste der Elemente
	 * @param visited Elemente die bereits in der Liste eingetragen wurden.
	 */
	protected void getAttributeOrder(final ArrayList<AttributeDescription> list, final HashSet<AttributeDescription> visited) {
		if(left != null) left.getAttributeOrder(list, visited);
		if(right != null) right.getAttributeOrder(list, visited);
	}

	/**
	 * Wertet Baum aus.
	 *
	 * @param realElements      Mapping der realen Elemente zu den ihnen zugeordeten Werten.
	 * @param tempElements      Mapping der temporären Attribute zu den ihnen zugeordneten Werten.
	 * @param pseudoObjectIndex Falls ein Attribut Aliase verwendet kann hier der Pseudoobjekt-Index festgelegt werden, von welchem Pseudoobjekt/Alias der Wert
	 *                          geholt werden soll. Ansonsten -1.
	 *
	 * @return Wert, den dieser Teilbaum speichert
	 * 
	 * @deprecated {@link #evaluateEx(Hashtable, Hashtable, int)} liefert zusätzlich den Datenzustand zurück
	 */
	@Deprecated
	public final ExpressionResult evaluate(
			final Hashtable<RealElement, BaseDataSet> realElements, final Hashtable<TempAttributeDescription, TempAttribut> tempElements, final int pseudoObjectIndex) {
		return evaluateEx(realElements, tempElements, pseudoObjectIndex).getResult();
	}
	
	/**
	 * Wertet Baum aus.
	 *
	 * @param realElements      Mapping der realen Elemente zu den ihnen zugeordeten Werten.
	 * @param tempElements      Mapping der temporären Attribute zu den ihnen zugeordneten Werten.
	 * @param pseudoObjectIndex Falls ein Attribut Aliase verwendet kann hier der Pseudoobjekt-Index festgelegt werden, von welchem Pseudoobjekt/Alias der Wert
	 *                          geholt werden soll. Ansonsten -1.
	 *
	 * @return Wert, den dieser Teilbaum speichert
	 */
	public ExpressionResultAndState evaluateEx(
			final Hashtable<RealElement, BaseDataSet> realElements,
			final Hashtable<TempAttributeDescription, TempAttribut> tempElements,
			final int pseudoObjectIndex) {
		ExpressionResultAndState leftR = null;
		ExpressionResultAndState rightR = null;
		ExpressionResultAndState single = null;

		final ResultType type;

		//falls nur ein Kind einen Wert hat, benutze 'single'
		if(left != null && right != null) {
			leftR = left.evaluateEx(realElements, tempElements, pseudoObjectIndex);
			rightR = right.evaluateEx(realElements, tempElements, pseudoObjectIndex);
			
			if(!operation.isBinary()){
				throw new IllegalStateException("Unärer Operator zusammen mit 2 Parametern verwendet");
			}

			return evaluateBinaryExpression(leftR, rightR);

		}
		else if(left != null) {
			single = left.evaluateEx(realElements, tempElements, pseudoObjectIndex);
			if(operation.isBinary()){
				throw new IllegalStateException("Binärer Operator zusammen mit einem Parameter verwendet");
			}			
			type = single.getType();
			return evaluateUnaryExpression(single, type);
		}
		else {
			throw new IllegalStateException("ExpressionTree nicht initialisiert");
		}
	}

	private ExpressionResultAndState evaluateBinaryExpression(final ExpressionResultAndState leftResultAndState, final ExpressionResultAndState rightResultAndState) {
		ResultType leftType = leftResultAndState.getType();
		ResultType rightType = rightResultAndState.getType();
		ExpressionResult leftResult = leftResultAndState.getResult();
		ExpressionResult rightResult = rightResultAndState.getResult();
		//do boolean operation
		assert operation.isBinary();

		// Zahl operator Zahl
		if((leftType == ResultType.DOUBLE || leftType == ResultType.LONG) && (rightType == ResultType.LONG || rightType == ResultType.DOUBLE)) {
			final ArithmeticOperationEx v = new ArithmeticOperationEx(leftResultAndState, rightResultAndState);
			v.setLeftTree(left);
			v.setRightTree(right);
			switch(operation) {
				case mult:
					return v.performMultOperation();
				case div:
					return v.performDivOperation();
				case divInt:
					return v.performDivIntOperation();
				case modulo:
					return v.performModOperation();
				case add:
					return v.performPlusOperation();
				case subtract:
					return v.performMinusOperation();
				case less:
					return v.less();
				case lessEqual:
					return v.lessEqual();
				case greater:
					return v.greater();
				case greaterEqual:
					return v.greaterEqual();
				case equal:
					return v.equal();
				case notEqual:
					return v.notEqual();
			}
		}
		// Zahl vergleichsOperator Status
		// Status vergleichsOperator Zahl
		else if(((leftType == ResultType.DOUBLE || leftType == ResultType.LONG) && rightType == ResultType.STATUS)
				|| (leftType == ResultType.STATUS && (rightType == ResultType.LONG || rightType == ResultType.DOUBLE))) {
			switch(operation) {
				case less:
				case lessEqual:
				case greater:
				case greaterEqual:
				case equal:
					return deriveResult(false, leftResultAndState, rightResultAndState);
				case notEqual:
					return deriveResult(true, leftResultAndState, rightResultAndState);
			}
		}
		// bool = bool
		else if((leftType == ResultType.BOOL && rightType == ResultType.BOOL) && operation == Operation.equal) {
			return deriveResult(leftResult.getBool() == rightResult.getBool(), leftResultAndState, rightResultAndState);
		}
		// bool <> bool
		else if((leftType == ResultType.BOOL && rightType == ResultType.BOOL) && operation == Operation.notEqual) {
			return deriveResult(leftResult.getBool() != rightResult.getBool(), leftResultAndState, rightResultAndState);
		}
		// bool und/oder bool
		// bool und/oder fehler
		// fehler und/oder bool
		else if((leftType == ResultType.BOOL && rightType == ResultType.BOOL) || (leftType == ResultType.BOOL && rightType == ResultType.ERROR) || (leftType == ResultType.ERROR && rightType == ResultType.BOOL) ) {
			switch(operation) {
				case and:
					return deriveResult(leftResult.getBool() && rightResult.getBool(), leftResultAndState, rightResultAndState);
				case or:
					return deriveResult(leftResult.getBool() || rightResult.getBool(), leftResultAndState, rightResultAndState);
			}
		}
		// leer op egal
		// egal op leer
		else if(leftType == ResultType.NONE || rightType == ResultType.NONE) {
			return ExpressionResultAndState.noResult(leftResultAndState, rightResultAndState);
		}
		// string op{& = <>} egal
		// status op{& = <>} egal
		// egal op{& = <>} string
		// egal op{& = <>} status
		else if(leftType == ResultType.STRING || rightType == ResultType.STRING || leftType == ResultType.STATUS || rightType == ResultType.STATUS) {
			switch(operation) {
				case conc:
					return deriveResult(leftResult.getString() + rightResult.getString(), leftResultAndState, rightResultAndState);
				case equal:
					return deriveResult(leftResult.getString().equals(rightResult.getString()), leftResultAndState, rightResultAndState);
				case notEqual:
					return deriveResult(!leftResult.getString().equals(rightResult.getString()), leftResultAndState, rightResultAndState);
			}
		}

		return ExpressionResultAndState.error(ErrorMessageBuilder.illegalOperation(operation, left, right, leftResultAndState, rightResultAndState));
	}

	private ExpressionResultAndState evaluateUnaryExpression(final ExpressionResultAndState param, final ResultType paramType) {

		assert !operation.isBinary();

		if(paramType == ResultType.BOOL) {
			if(operation == Operation.not) {
				return deriveResult(!(param.getResult().getBool()), param);
			}
		}
		else if(paramType == ResultType.DOUBLE || paramType == ResultType.LONG) {
			if(operation == Operation.neg) {
				return deriveResult(-1 * param.getResult().getDouble(), param);
			}
		}
		else if(paramType == ResultType.NONE) {
			return ExpressionResultAndState.noResult(param);
		}
		return ExpressionResultAndState.error(ErrorMessageBuilder.illegalOperation(operation, left, param));
	}

	/** Gibt den erwarteten Ergebnis-Typ der Operation zurück
	 * @param ast*/
	public ResultType getResultType(final AST ast) throws SemanticErrorException {
		ResultType leftType = null;
		ResultType rightType = null;
		ResultType singleType = null;

		final ResultType type;

		//falls nur ein Kind einen Wert hat, benutze 'single'
		if(left != null && right != null) {
			leftType = left.getResultType(ast);
			rightType = right.getResultType(ast);
		}
		else if(right != null) {
			if(operation != Operation.neg && operation != Operation.not) {
				throw new SemanticErrorException(InterpreterErrorMessage.INVALID_EXPRESSION + this + " (Operation benötigt zwei Operatoren)", ast);
			}
			else {
				singleType = right.getResultType(ast);
			}
		}
		else if(left != null) {
			if(operation != Operation.neg && operation != Operation.not) {
				throw new SemanticErrorException(InterpreterErrorMessage.INVALID_EXPRESSION + this + " (Operation benötigt zwei Operatoren)", ast);
			}
			else {
				singleType = left.getResultType(ast);
			}
		}
		else {
			throw new SemanticErrorException(InterpreterErrorMessage.INVALID_EXPRESSION + this + " (Operation benötigt Operatoren)", ast);
		}

		//Typ setzen
		if(singleType == null) {
			if(rightType != leftType) {
				//wenn ein double Wert vorkommt, dann ist das Ergebnis wieder ein double
				//OPT sollte in ArithmeticOperation gesetzt werden.
				if((rightType == ResultType.LONG && leftType == ResultType.DOUBLE) || (rightType == ResultType.DOUBLE && leftType == ResultType.LONG)) {
					type = ResultType.DOUBLE;
				}
				else if(rightType == ResultType.STRING || leftType == ResultType.STRING) {
					type = ResultType.STRING;
				}				
				else {
					throw new SemanticErrorException(
							InterpreterErrorMessage.INVALID_EXPRESSION + this + " (Inkompatible Typen: " + leftType + "; " + rightType + ")", ast
					);
				}
			}
			else {
				type = rightType;
			}
		}
		else {
			type = singleType;
		}


		if(type == ResultType.BOOL) {
			return ResultType.BOOL;
		}
		//arithmetical op
		else if(type == ResultType.DOUBLE || type == ResultType.LONG) {
			if(operation == Operation.neg && type == ResultType.DOUBLE) {
				return ResultType.DOUBLE;
			}
			else if(operation == Operation.neg) {
				return ResultType.LONG;
			}
			else	//operation with two children
			{
				if(operation == Operation.mult) {
					return type;
				}
				else if(operation == Operation.div) {
					return ResultType.DOUBLE;
				}
				else if(operation == Operation.divInt) {
					return ResultType.LONG;
				}
				else if(operation == Operation.modulo) {
					return type;
				}
				else if(operation == Operation.add) {
					return type;
				}
				else if(operation == Operation.subtract) {
					return type;
				}
				else if(operation == Operation.less) {
					return ResultType.BOOL;
				}
				else if(operation == Operation.lessEqual) {
					return ResultType.BOOL;
				}
				else if(operation == Operation.greater) {
					return ResultType.BOOL;
				}
				else if(operation == Operation.greaterEqual) {
					return ResultType.BOOL;
				}
				else if(operation == Operation.equal) {
					return ResultType.BOOL;
				}
				else if(operation == Operation.notEqual){
					return ResultType.BOOL;
				}
			}
		}
		//String Ops
		else if(type == ResultType.STRING || type == ResultType.STATUS) {
			if(operation == Operation.conc) {
				return ResultType.STRING;
			}
			else if(operation == Operation.equal) {
				return ResultType.BOOL;
			}
			else if(operation == Operation.notEqual){
				return ResultType.BOOL;
			}
		}
		else if(type == ResultType.NONE){
			return ResultType.NONE;
		}
		throw new SemanticErrorException(InterpreterErrorMessage.INVALID_EXPRESSION + this + " (Operator ist nicht definiert für: " + type + ")", ast);
	}

	/** Gibt zurück, für wie viele Spalten ein TempAttribut definiert werden soll, dass diesen Ausdruck verwendet (bei der Verwendung von Aliasen in Attributen)
	 * @return Anzahl der Spalten
	 */
	public int getNumColumns() {
		int columns = 1;
		if(left != null) {
			columns = Math.max(columns, left.getNumColumns());
		}
		if(right != null)  {
			columns = Math.max(columns, right.getNumColumns());
		}
		return columns;
	}

	/** @see java.lang.Object#toString() */
	@Override
    public String toString() {
		if(left != null && right != null) {
		    return ("(" + left + operation.getSymbol() + right + ")");
		}
		else {
			return operation.getSymbol() + left;
		}
	}

	/**
	 * Liefert das erste Reale Attribut, das im Ausdruck verwendet wird.
	 *
	 * @return Attributbeschreibung oder null, falls kein Reales Attribut verwendet wird.
	 */
	public RealAttributeDescription getFirstRealAttribute() {
		RealAttributeDescription rad = null;
		if(left != null) rad = left.getFirstRealAttribute();
		if(rad == null && right != null) rad = right.getFirstRealAttribute();
		return rad;
	}

	/**
	 * Liefert das erste temporäre Attribut, das im Ausdruck verwendet wird.
	 *
	 * @return Attributbeschreibung oder null, falls kein temporäres Attribut verwendet wird.
	 */
	public TempAttributeDescription getFirstTempAttribute() {
		TempAttributeDescription tad = null;
		if(left != null) tad = left.getFirstTempAttribute();
		if(tad == null && right != null) tad = right.getFirstTempAttribute();
		return tad;
	}

	/**
	 * Liefert ein Set mit den temporären Attributen, die dieser Ausdruck direkt verwendet. Dazu wird der Ausdruck durchsucht und die gefundenen temporären
	 * Attribute zurückgegeben. Die gefundenen Attribute werde nicht rekursiv durchsucht.
	 *
	 * @return Temporäre Attribute
	 */
	public HashSet<TempAttributeDescription> getDirectTempAttributes() {
		final HashSet<TempAttributeDescription> tads = new HashSet<TempAttributeDescription>();
		if(left != null) tads.addAll(left.getDirectTempAttributes());
		if(right != null) tads.addAll(right.getDirectTempAttributes());

		return tads;
	}

	/**
	 * Liefert alle realen Attribute die im Ausdruck verwendet werden. Es werden nur direkt referenzierte reale Attribute zurückgegeben. D.h. temporäre Attribute
	 * und deren reale Attribute werden nicht zurückgegeben.
	 *
	 * @return Set mit realen Attributen.
	 */
	public HashSet<RealAttributeDescription> getRealAttributes() {
		final HashSet<RealAttributeDescription> realAttributes = new HashSet<RealAttributeDescription>();

		if(left != null) realAttributes.addAll(left.getRealAttributes());

		if(right != null) realAttributes.addAll(right.getRealAttributes());

		return realAttributes;
	}

	/** @see java.lang.Object#equals(java.lang.Object) */
	@Override
    public boolean equals(final Object other) {
		boolean result = false;
		if(other != null && other instanceof ExpressionTree) {
			final ExpressionTree tree = (ExpressionTree)other;
			if(left != null) result = left.equals(tree.left);
			if(right != null) result &= right.equals(tree.right);
			result &= operation == tree.operation;
		}
		return result;
	}

	/**
	 * Überprüft ob der Ausdruck ein Attribut enthält. Es wird nach dem ersten temporären oder realen Attribut gesucht.
	 *
	 * @return	<code>true</code> falls der Ausdruck ein Attribut enthält.<code>false</code> sonst.
	 */
	public boolean containsAttribute() {
		boolean result = false;
		if(left != null) result |= left.containsAttribute();
		if(right != null) result |= right.containsAttribute();
		return result;
	}
}
