/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.util;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.pua.prot.aggregations.AbstractAggregation;
import de.bsvrz.pua.prot.aggregations.AggregationResultValue;
import de.bsvrz.pua.prot.util.ExpressionResult.ResultType;
import de.bsvrz.sys.funclib.losb.datk.PidMisc;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Objects;
import java.util.function.Supplier;

/**
 * Enthält einen Wert für eine Spalte des Protokolldatensatzes, d.h. den Wert eines realen oder temporären Attributs oder einer Attributgruppe. <br> Wichtig:
 * <br> 1. Unterstützt keine Bitfelder! Diese werden von Dem K2S Serialisierer nicht unterstützt!<br> 2. Grund, warum keine Data Objekte versandt werden.
 * Seitens der Kernsoftware ist es nur vorgesehen, Data Objekte zu übertragen, die eine Attributgruppe abbilden. Teile einer Attributgruppe können nicht
 * übertragen werden. Aber gerade dies wird für PuA benötigt - da man ja auch einzelne Attributwerte übertragen soll!
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class ResultValue implements Serializable {

	/**
	 * Operationen die auf ResultValue Objekten durchgeführt werden können.
	 *
	 * @author beck et al. projects GmbH
	 * @author Martin Hilgers
	 */
	@Deprecated
	public enum ResultValueOperation {

		/** Addition */
		PLUS,
		/** Multiplikation */
		MULT,
		/** Division */
		DIV,
		/** Subtraktion */
		MINUS
	}

	;

	private static final long serialVersionUID = -273122742794554533L;

	/** Wird für Referenzen vergeben, denen kein Objekt zugeordnet ist. */
	public static final long INVALID_ID = -1;

	/** Wert, falls ResultValue atomar. */
	protected ExpressionResult _value = null;

	/** Kinder, falls ResultValue nicht-atomares Element repräsentiert. (Listen, Felder, Attributgruppen, etc). */
	protected List<ResultValue> _children = null;

	/** Zeigt an ob das gespeicherte Datum aggregiert werden darf. (Siehe TAnf 1.1, S. 22 Integerwerte bei Aggregationen) */
	private boolean _isSummable = true;

	/**
	 * Erzeugt Objekt für reales oder temporäres Attribut oder eine Attributgruppe. Holt den Wert aus dem <code>Data</code> Objekt.
	 *
	 * @param data Daten.
	 */
	public ResultValue(Data data) {
		storeData(data);
	}

	public ResultValue(final List<ResultValue> children) {
		for(final ResultValue child : children) {
			if(child == null) throw new NullPointerException();
		}
		this._children = children;
	}

	/** @param expressionResult Ergebnis eines Ausdrucks */
	public ResultValue(ExpressionResult expressionResult) {
		this._value = expressionResult;
	}

	/** @return Zeigt an ob nur ein Wert gespeichert wird, oder mehrere. */
	public final boolean isAtomar() {
		return _value != null;
	}

	/**
	 * Zeigt an ob Daten vorhanden sind.
	 *
	 * @return <code>true</code> falls Daten abholbereit sind.
	 */
	public final boolean hasData() {
		return _value != null || (_children != null && !_children.isEmpty());
	}

	/** @return Liefert den Wert zurück, falls es sich um ein einfaches Attribut handelt. */
	public final ExpressionResult getValue() {
		return _value;
	}

	/**
	 * Legt Daten in dem Objekt ab. Falls <code>data</code> nicht atomar ist, werden die Subdaten ebenfalls abgelegt.
	 *
	 * @param data
	 */
	private void storeData(Data data) {
		if(data == null) {
			_value = new ExpressionResult();
		}
		else {
			if(data.isPlain()) {
				_value = new ExpressionResult();
				AttributeType type = data.getAttributeType();
				if(type instanceof StringAttributeType) // String
				{
					_value.set(data.asTextValue().getText());
				}
				else if(type instanceof IntegerAttributeType) // Integer oder Bool
				{
					if(type.getPid().equals(PidMisc.PID_BOOLEAN_TYPE)) {
						byte bv = data.asUnscaledValue().byteValue();
						if(bv == 0) {
							_value.set(false);
						}
						else if(bv == 1) {
							_value.set(true);
						}
						else {
							_value.setError("Ungültiger boolscher Wert.");
						}
					}
					else {
						// Integer Wert setzten
						long x = data.asUnscaledValue().longValue();
						// Es muss festgestellt werden,
						// ob es sich um einen Wert außerhalb des Wertebereichs handelt.
						// dieser darf nicht aggregiert werden.
						IntegerAttributeType iat = (IntegerAttributeType)data.getAttributeType();

						final IntegerValueRange range = iat.getRange();
						if(range != null) {
							_value.setUnscaledValue(x, range.getConversionFactor());

							if(x < iat.getRange().getMinimum() || x > iat.getRange().getMaximum())
							// Wert darf nicht zum aggregieren verwendet werden
							{
								_isSummable = false;
							}
						}
						else{
							_value.setUnscaledValue(x, 1.0);
							_isSummable = false;
						}

						final IntegerValueState state = getState(x, iat.getStates());
						if(state != null){
							_value.setState(state.getName(), _isSummable);
						}
					}
				}
				else if(type instanceof DoubleAttributeType) // Double
				{
					_value.set(data.asScaledValue().doubleValue());
				}
				else if(type instanceof TimeAttributeType) // Time
				{
					_value.setUnscaledValue(data.asTimeValue().getMillis());
				}
				else if(type instanceof ReferenceAttributeType) // Referenz
				{
					if(data.asReferenceValue().getSystemObject() != null) {
						_value.setUnscaledValue(data.asReferenceValue().getSystemObject().getId());
					}
					else {
						_value.setUnscaledValue(INVALID_ID);
					}
				}
				else {
					_value = ExpressionResult.error(
							"Unbekannter Attributtyp kann nicht verwendet werden: " + type + ". Aufgetreten in " + data
					);
				}
			}
			else {
				// nicht atomar -> Liste oder Feld
				if(data.isArray()) {
					_children = new ArrayList<ResultValue>(data.asArray().getLength());
				}
				else {
					_children = new ArrayList<ResultValue>();
				}

				for(Iterator iterator = data.iterator(); iterator.hasNext();) {
					final Data innerData = (Data)iterator.next();
					_children.add(new ResultValue(innerData));
				}
			}
		}
	}

	private static IntegerValueState getState(final long x, final List<IntegerValueState> states) {
		for(final IntegerValueState state : states) {
			if(x == state.getValue()) return state;
		}
		return null;
	}

	/** @return Returns the children. */
	public List<ResultValue> getChildren() {
		return _children;
	}

	/**
	 * Verkleinert die Liste mit den Werten. Es werden nur die ersten <code>i</code> Werte behalten, alle weiteren werden verworfen.
	 *
	 * @param i Neue Länge. Muss zwischen <code>0</code> und <code>{@link #getChildren()}.size()</code> liegen.
	 */
	public void truncChildren(int i) {
		if(_children != null) {
			if(i < 0 || i > _children.size()) {
				throw new IllegalArgumentException(
						"Parameterwert außerhalb des Wertebereiches: [0, " + _children.size() + "]"
				);
			}
			else {
				_children = _children.subList(0, i);
			}
		}
	}

	/** @see java.lang.Object#toString() */
	@Override
    public String toString() {
		StringBuffer sb = new StringBuffer();

		if(this.isAtomar()) {
			sb.append(_value);
		}
		else {
			sb.append("[");
			final List<ResultValue> children = this.getChildren();
			if(children.size() > 0){
				for(final ResultValue rv : children) {
					sb.append(rv.toString()).append(", ");
				}
				sb.setLength(sb.length() - 2);
			}
			sb.append("]");
		}
		return sb.toString();
	}

	/** @see java.lang.Object#equals(java.lang.Object) */
	@Override
    public boolean equals(Object object) {
		if(object == null) {
			return false;
		}
		else if(object instanceof ResultValue) {
			ResultValue other = (ResultValue)object;
			if(this.isAtomar() && other.isAtomar()) // beide atomar
			{
				return this._value.equals(other._value);
			}
			else if(!this.isAtomar() && !other.isAtomar() && compareChildren(other)) {
				return true;
			}
			else {
				return false;
			}
		}
		else {
			return false;
		}
	}

	@Override
	public int hashCode() {
		return Objects.hash(_children, _value);
	}

	/**
	 * Vergleicht <code>children</code> von <code>this</code> und <code>other</code>
	 *
	 * @param other Vergleichsobjekt.
	 *
	 * @return <code>true</code> falls die Objekte gleich sind. <code>false</code> sonst.
	 */
	private boolean compareChildren(ResultValue other) {
		if(this._children.size() == other._children.size()) {
			boolean result = true;
			for(int i = 0; i < this._children.size() & result; i++) {
				result &= this._children.get(i).equals(other._children.get(i));
			}
			return result;
		}
		else {
			return false;
		}
	}

	/**
	 * Führt eine arithmetische Operation durch. <code>this</code> ist dabei der linke Operand. Die beiden Objekte müssen die gleiche Struktur haben. Kann die
	 * Operation nicht angewendet werden, so wird der Wert des Rückgabewerts auf {@link de.bsvrz.pua.prot.util.ExpressionResult.ResultType#NONE} gesetzt.
	 *
	 * @param other     Rechter Operand.
	 * @param operation Operation.
	 *
	 * @return <code>this</code>
	 * 
	 * @deprecated Diese Methode wird nicht mehr aufgerufen und in Zukunft entfernt.
	 */
	@Deprecated
	public ResultValue performOperation(ResultValue other, ResultValueOperation operation) {
		if(other == null) {
			return this;
		}
		else {
			traverse(this, other, operation);
			return this;
		}
	}

	/**
	 * Führt die Operation für jedes Element der Objekte durch. Das Ergebnis wird im linken Operanden abgelegt.
	 *
	 * @param left  Linker Operand. Wird verändert.
	 * @param right Rechter Operand. Wird nicht geändert.
	 *
	 * @deprecated Diese Methode wird nicht mehr aufgerufen und in Zukunft entfernt.
	 */
	@Deprecated
	private static void traverse(ResultValue left, ResultValue right, ResultValueOperation operation) {
		if(left.isAtomar()) {
			if(!right.isAtomar()) {
				left.set(ExpressionResult.error());
			}
			else {
				left.set(performOperationOn(left.getValue(), right.getValue(), operation));
			}
		}
		else {
			if(right.isAtomar() | left.getChildren().size() != right.getChildren().size()) left.set(ExpressionResult.error());
			for(int i = 0; i < left.getChildren().size(); i++) {
				ResultValue.traverse(left.getChildren().get(i), right.getChildren().get(i), operation);
			}
		}
	}

	/**
	 * Setzt den Wert des Objekts.
	 *
	 * @param result
	 */
	public void set(ExpressionResult result) {
		this._value = result;
		this._children = null;
	}

	/**
	 * Führt eine arithmetische Operation auf den beiden Operanden aus.
	 *
	 * @param left      Linker Operand.
	 * @param right     Rechter Operand.
	 * @param operation Operation.
	 *
	 * @return Ergebnis der Operation.
	 *
	 * @deprecated Diese Methode wird nicht mehr aufgerufen und in Zukunft entfernt.
	 */
	@Deprecated
	private static ExpressionResult performOperationOn(ExpressionResult left, ExpressionResult right, ResultValueOperation operation) {
		ResultType leftType = left.getType();
		ResultType rightType = right.getType();
		if(leftType.equals(rightType) && (leftType == ResultType.DOUBLE || leftType == ResultType.LONG)) {
			ArithmeticOperation op = new ArithmeticOperation(left);
			op.set(right);
			if(operation == ResultValueOperation.PLUS) {
				return op.performPlusOperation();
			}
			else if(operation == ResultValueOperation.MULT) {
				return op.performMultOperation();
			}
			else if(operation == ResultValueOperation.DIV) {
				return op.performDivIntOperation();
			}
			else if(operation == ResultValueOperation.MINUS) {
				return op.performMinusOperation();
			}
			else {
				return ExpressionResult.error();
			}
		}
		else {
			return new ExpressionResult();
		}
	}

	/**
	 * Erstellt eine Kopie des Objekts, bei der für die Aggregation ungeeignete Werte den Typ {@link de.bsvrz.pua.prot.util.ExpressionResult.ResultType#NONE}
	 * erhalten.
	 *
	 * @return Kopie des Objekts.
	 *
	 * @throws CloneNotSupportedException Fehler beim Klonen.
	 * @see java.lang.Object#clone()
	 * @deprecated Eine bessere Alternavtive ist {@link AggregationResultValue#create(ResultValue, Supplier, AbstractAggregation)}
	 */
	@Deprecated
	public ResultValue cloneForAggregation() throws CloneNotSupportedException {
		ResultValue clone = new ResultValue();
		if(isAtomar()) {
			if(this.isSummable()) {
				clone._value = this._value.clone();
			}
			else {
				clone._value = new ExpressionResult();
			}
			clone._children = null;
		}
		else {
			clone._value = null;
			clone._children = new ArrayList<ResultValue>(_children.size());
			for(ResultValue rv : _children) {
				clone._children.add(rv.cloneForAggregation());
			}
		}
		return clone;
	}

	/**
	 * Zeigt an, ob der Wert in einer Aggregation verwendet werden darf.
	 *
	 * @return <code>true</code> falls der Wert von einer Aggregation verwendet werden darf, <code>false</code> sonst.
	 */
	public boolean isSummable() {
		return _isSummable;
	}

	/** Nötig zum Deserialisieren. Sollte nicht verwendet werden. */
	protected ResultValue() {
	}

	/**
	 * Serialisiert das Objekt. (ohne Namen, Pid, id und Flags isArray und isList)
	 *
	 * @param out
	 *
	 * @throws IOException
	 */
	private void writeObject(ObjectOutputStream out) throws IOException {
		if(_children != null) {
			out.writeBoolean(false);
			out.writeInt(_children.size());
			for(ResultValue child : _children) {
				child.writeObject(out);
			}
		}
		else if(_value != null) {
			out.writeBoolean(true);
			ResultType type = _value.getType();
			if(type == ResultType.BOOL) {
				out.writeByte(0);
				out.writeBoolean(_value.getBool());
			}
			else if(type == ResultType.DOUBLE) {
				out.writeByte(1);
				out.writeDouble(_value.getDouble());
			}
			else if(type == ResultType.ERROR) {
				out.writeByte(2);
				out.writeObject(_value.getString());
			}
			else if(type == ResultType.LONG) {
				out.writeByte(3);
				out.writeLong(_value.getLong());
			}
			else if(type == ResultType.NONE) {
				out.writeByte(4);
			}
			else if(type == ResultType.STRING) {
				out.writeByte(5);
				out.writeObject(_value.getString());
			}
			else if(type == ResultType.STATUS) {
				out.writeByte(6);
				out.writeObject(_value.getString());
			}
			else if(type == ResultType.NO_CHANGE) {
				out.writeByte(7);
			}
		}
	}

	/**
	 * Deserialisiert das Objekt (soweit möglich)
	 *
	 * @param in
	 *
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
		if(!in.readBoolean()) {
			// read child-objects
			int countChildren = in.readInt();
			this._children = new ArrayList<ResultValue>(countChildren);
			for(int i = 0; i < countChildren; i++) {
				ResultValue rv = new ResultValue();
				rv.readObject(in);
				this._children.add(rv);
			}
		}
		else {
			// read value
			this._value = new ExpressionResult();
			switch(in.read()) {
				case 0:
					this._value.set(in.readBoolean());
					break;
				case 1:
					this._value.set(in.readDouble());
					break;
				case 2:
					this._value.setError((String)in.readObject());
					break;
				case 3:
					this._value.setUnscaledValue(in.readLong(), 1);
					break;
				case 4:
					break;
				case 5:
					this._value.set((String)in.readObject());
					break;
				case 6:
					this._value.setState((String)in.readObject(), false);
					break;
				case 7:
					this._value.setNoChange();
					break;
			}
		}
	}
}
